/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: interpol.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: interpol.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
/*---------------------------------------------------------------------- -*/
/*    Utility module for interpolating in arrays                          */
/*----------------------------------------------------------------------- */
#include <math.h>
#include "interpol.h"

/* file variables */
float xint[10], yint[10]; /* interpolation arrays */
float denom[10]; /* Lagrangian denominators */
int   ixcen=0, iycen=0;   /* first element to consider */

/* internal prototypes */
void iminfn (float pos, int n, int ihalf, float *intr, int *cen);

float imintp (float *array, int nx, int ny, int half, 
	      float xpix, float ypix, float fblank)
/*----------------------------------------------------------------------- */
/*   Interpolate value.  Returned value blanked if outside image or most  */
/*   data blanked.  Arguments are in pixels.                              */
/*   Inputs:                                                              */
/*      array    data array to be interpolated                            */
/*      nx       dimension of array in "X" (fastest dimension)            */
/*      ny       dimension of array in "Y" (slowest varying dimension)    */
/*      half     interpolation half width in range (1,4)                  */
/*      xpix     pixel location (0-rel) on X axis to interpolate at       */
/*      ypix     pixel location (0-rel) on Y axis to interpolate at       */
/*      fblank   magic blanking value for array                           */
/*   Returns:                                                             */
/*      Interpolated value.                                               */
/*----------------------------------------------------------------------- */
{
  int   i, j, k, good, iwid, ihalf;
  long  iap;
  float sum, sumwt, wt, wty, row[10], prod, den, xp, yp;
  float val; /* return value */

  val = fblank;
  
  /* Outside array? */
  if ((xpix<0.0) || (ypix<0.0) ||
      (xpix>nx-0.5) || (ypix>ny-0.5)) return val;
  
  /* interpolation kernal width */
  iwid = half * 2 + 1;
  if (iwid<3) iwid = 3;
  if (iwid>9) iwid = 9;
  ihalf = iwid / 2;

  /* Get interpolation fn */
  if (xpix!=xint[0]) iminfn (xpix, nx, ihalf, xint, &ixcen);
  if (ypix!=yint[0]) iminfn (ypix, ny, ihalf, yint, &iycen);

  sum = 0.0;
  sumwt = 0.0;
  good = 0;
  for (j=0; j<iwid; j++) {
    wty = yint[j+1];
    for (i=0; i<iwid; i++) {
      /* data address */
      iap = ixcen + i + ((iycen + j) * nx);
      wt = xint[i+1] * wty;
      if (array[iap]!=fblank) {
	sumwt = sumwt + wt;
	sum = sum + array[iap] * wt;
	good = good + 1;
      }
    } /* end i loop */
  } /* end j loop */
      
  /* Normalize sum */
  if (sumwt>0.90) { /* done if sufficient weight */
    val = sum / sumwt;
    return val;
  }

  /* Too much blanked data; try again using knowledge of blanks if */
  /* sufficient data. */
  if (good<(iwid * iwid)/3) return fblank; /* don't bother */

  /* Set "x" at first pixel to 1.0 */
  xp = xpix - ixcen;
  yp = ypix - iycen;
  for (j=0; j<iwid; j++) {
    /* Interpolate in rows */
    sum = 0.0;
    sumwt = 0.0;
    for (i=0; i<iwid; i++) {
      den = 1.0;
      prod = 1.0;
      for (k=0; k<iwid; k++) {
	iap = ixcen + k + ((iycen + j) * nx);
	if (array[iap]!=fblank) {
	  if (i!=k) {
	    den *=  (float)(i - k);
	    prod *= (float)(xp - k);
	  }
	}
      } /* end k loop */

     /*  Accumulate */
      iap = ixcen + i + ((iycen + j) * nx);
      if (array[iap]!=fblank) {
	if (abs (den)>1.0e-10) {
	  wt = prod / den;
	} else {
	  wt = 0.0;
	}
	sumwt = sumwt + wt;
	sum = sum + wt * array[iap];
      }
    } /* end i loop */

    /* Interpolate column value */
    if (sumwt>0.5) {
      row[j] = sum / sumwt;
    } else {
      row[j] = fblank;
    }
  } /* end j loop */
  
  /* Interpolate in column */
  sum = 0.0;
  sumwt = 0.0;
  for (i=0; i<iwid; i++) {
    den = 1.0;
    prod = 1.0;
    for (k=0; k<iwid; k++) {
      if (row[k]!=fblank) {
	if (i!=k) {
	  den  *= (float)(i - k);
	  prod *= (float)(yp - k);
	}
      }
    } /* end k loop */

    /* Accumulate */
    if (row[i]!=fblank) {
      if (abs (den)>1.0e-20) {
	wt = prod / den;
      } else {
	wt = 0.0;
      }
      sumwt = sumwt + wt;
      sum = sum + wt * row[i];
    }
  } /* end i loop */

  /*  Interpolate value */
  if (sumwt>0.5) {
    val = sum / sumwt;
  } else {
    val = fblank;
  }
  return val;
} /* end imintp */

void cxintp (float cxout[2], float *array, int nx, int half, 
	      float xpix, float fblank)
/*----------------------------------------------------------------------- */
/*   Interpolate complex value.  Returned values blanked if outside array */
/*   or most data blanked.  Arguments are in pixels.                      */
/*   Inputs:                                                              */
/*      array    data array to be interpolated as (real, imag)            */
/*      nx       dimension of array.                                      */
/*      half     interpolation half width in range (1,4)                  */
/*      xpix     pixel location (0-rel) to interpolate at                 */
/*      fblank   magic blanking value for array                           */
/*   Outputs:                                                             */
/*      cxout    returned complex value as real, imaginary                */
/*----------------------------------------------------------------------- */
{
  int   i, j, k, iwid, ihalf;
  long  iap;
  float sumre, sumim, sumwt, wt;

  /* initialize results */
  cxout[0] = fblank;
  cxout[1] = fblank;
  
  /* Outside array? */
  if ((xpix<0.0) || (xpix>nx-0.5)) return;
  
  /* interpolation kernal width */
  iwid = half * 2 + 1;
  if (iwid<3) iwid = 3;
  if (iwid>9) iwid = 9;
  ihalf = iwid / 2;

  /* Get interpolation fn */
  if (xpix!=xint[0]) iminfn (xpix, nx, ihalf, xint, &ixcen);

  sumre = sumim = 0.0;
  sumwt = 0.0;
  for (i=0; i<iwid; i++) {
    /* data address - as complex*/
    iap = 2 * (ixcen + i);
    wt = xint[i+1];
    if ((array[iap]!=fblank) && (array[iap+1]!=fblank)) {
      sumwt = sumwt + wt;
      sumre += array[iap] * wt;
      sumim += array[iap+1] * wt;
    }
  } /* end i loop */
      
  /* Normalize sum if sufficient weight*/
  if (sumwt>0.90) { 
    cxout[0] = sumre / sumwt;
    cxout[1] = sumim / sumwt;
  }
} /* end cxintp */


void iminfn (float pos, int n, int ihalf, float *intr, int *cen)
/*----------------------------------------------------------------------- */
/*   Returns Lagrangian interpolation weights.  Adjustments are made to   */
/*   account for the ends of the data.                                    */
/*   Inputs:                                                              */
/*      pos      Coordinate to interpolate at on axis                     */
/*      n        Maximum pixel number on axis                             */
/*      ihalf    half width in pixels of interpolation kernal             */
/*   Output:                                                              */
/*      intr     Interpolation array                                      */
/*                (1) = pos                                               */
/*                (2...) interpolation weights.                           */
/*      cen      (First - 1) pixel number for use                         */
/*----------------------------------------------------------------------- */
{
  int   ipos, i, j, iwid;
  float prod, xx;

  iwid = ihalf * 2 + 1;
  /* Calculate Lagrangian denominators if necessary */
  if (denom[0]!=(float)iwid) {
    denom[0] = iwid;
    for (j=0; j<iwid; j++) {
      prod = 1.0;
      for (i=0; i<iwid; i++) {
	if (i!=j) prod *= (float)(j - i);
      } /* end i loop */
      denom[j] = 1.0 / prod;
    } /* end j loop */
  } /* end compute  Lagrangian denominators */
  
  intr[0] = pos;

  /* fractional pixel */
  ipos = pos + 0.5;
  
  /* Set first pixel */
  *cen = ipos - ihalf;
  if (*cen<0) *cen = 0;
  if (*cen>=(n-iwid)) *cen = (n-iwid);

  xx = pos - *cen;
  /*  compute interpolating kernal */
  for (j=0; j<iwid; j++) {
    prod = denom[j];
    for (i=0; i<iwid; i++) {
      if (i!=j) prod *=  (xx - i);
    } /* end i loop */
    intr[j+1] = prod;
  } /* end j loop */
} /* end iminfn */

/* */
